/**
 * 
 */
package com.wieseke.cptk.statistics;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;

import com.wieseke.cptk.common.api.IHostNode;
import com.wieseke.cptk.common.api.IParasiteNode;
import com.wieseke.cptk.common.api.IPositionNode;
import com.wieseke.cptk.common.constants.ViewerConstants;
import com.wieseke.cptk.common.dao.ParasiteRank;
import com.wieseke.cptk.common.util.Tupel;
import com.wieseke.cptk.input.dao.InputCophylogeny;
import com.wieseke.cptk.input.dao.InputHostNode;
import com.wieseke.cptk.input.dao.InputNode;
import com.wieseke.cptk.input.dao.InputParasiteNode;
import com.wieseke.cptk.reconstruction.constants.BaseSolverConstants;
import com.wieseke.cptk.reconstruction.constants.StatisticConstants;

/**
 * @author Nicolas Wieseke
 * 
 */
public class TreeGenerator {

	private Random random;
	private String model;
	private Map<Integer, List<Double>> probabilityDistributions;
	private List<InputHostNode> hostLeaves;
	private List<InputParasiteNode> parasiteLeaves;

	public TreeGenerator(Random random) {
		if (random == null) {
			this.random = new Random();
		} else {
			this.random = random;
		}
		this.probabilityDistributions = new HashMap<Integer, List<Double>>();
		this.model = StatisticConstants.MODEL_MARKOV;
	}

	/**
	 * @param filename
	 * @param hostLeafNumber
	 * @param parasiteLeafNumber
	 * @param beta
	 * @param association
	 * @return
	 */
	public InputCophylogeny getRandomInputCophylogeny(String filename,
			int hostLeafNumber, int parasiteLeafNumber, String model,
			int association) {
		this.model = model;
		boolean isBetaSplitModel = false;
		boolean isAgeModel = false;
		if (model == StatisticConstants.MODEL_AGE)
			isAgeModel = true;
		else if (model == StatisticConstants.MODEL_MARKOV
				|| model == StatisticConstants.MODEL_ALDOUS
				|| model == StatisticConstants.MODEL_PDA)
			isBetaSplitModel = true;

		InputCophylogeny cophylogeny = InputCophylogeny
				.createNewInputCophylogeny(filename);
		this.hostLeaves = new ArrayList<InputHostNode>();
		this.parasiteLeaves = new ArrayList<InputParasiteNode>();

		// create trees
		if (isBetaSplitModel) {
			createRandomTreeByBetaSplitting(cophylogeny, hostLeafNumber,
					ViewerConstants.TREE_SELECTION_HOST);
			createRandomTreeByBetaSplitting(cophylogeny, parasiteLeafNumber,
					ViewerConstants.TREE_SELECTION_PARASITE);
		} else if (isAgeModel) {
			createRandomTreeByAge(cophylogeny, hostLeafNumber,
					ViewerConstants.TREE_SELECTION_HOST);
			createRandomTreeByAge(cophylogeny, hostLeafNumber,
					ViewerConstants.TREE_SELECTION_PARASITE);
		}

		cophylogeny.getHostRoot().renumberNodes(0);
		cophylogeny.getHostRoot().renameNodes();
		cophylogeny.getParasiteRoot().renumberNodes(0);
		cophylogeny.getParasiteRoot().renameNodes();

		// create associations
		switch (association) {
		case StatisticConstants.ASSOCIATION_UNIQUE:
			AssociationGenerator.createUniqueAssociations(hostLeaves,
					parasiteLeaves, this.random);
			break;
		case StatisticConstants.ASSOCIATION_MULTIPLE:
			AssociationGenerator.createMultipleAssociations(hostLeaves,
					parasiteLeaves, this.random);
			break;
		}

		// create ranks
		if (isBetaSplitModel) {
			createRandomRanks(cophylogeny, ViewerConstants.TREE_SELECTION_BOTH);
		} else if (isAgeModel)
			createRandomRanks(cophylogeny,
					ViewerConstants.TREE_SELECTION_PARASITE);

		return cophylogeny;
	}

	public InputCophylogeny getRandomInputCophylogeny(String filename,
			InputCophylogeny originalCophylogeny, String model, int tree) {
		this.model = model;
		boolean isBetaSplitModel = false;
		boolean isAgeModel = false;
		if (model == StatisticConstants.MODEL_AGE)
			isAgeModel = true;
		else if (model == StatisticConstants.MODEL_MARKOV
				|| model == StatisticConstants.MODEL_ALDOUS
				|| model == StatisticConstants.MODEL_PDA)
			isBetaSplitModel = true;

		InputCophylogeny cophylogeny = InputCophylogeny
				.createNewInputCophylogeny(filename);
		this.hostLeaves = new ArrayList<InputHostNode>();
		this.parasiteLeaves = new ArrayList<InputParasiteNode>();
		ArrayList<InputHostNode> originalHostLeaves = new ArrayList<InputHostNode>();
		ArrayList<InputParasiteNode> originalParasiteLeaves = new ArrayList<InputParasiteNode>();
		for (IPositionNode hostNode : originalCophylogeny
				.getNodeList(originalCophylogeny.getHostRoot())) {
			if (hostNode.isLeaf())
				originalHostLeaves.add((InputHostNode) hostNode);
		}
		for (IPositionNode parasiteNode : originalCophylogeny
				.getNodeList(originalCophylogeny.getParasiteRoot())) {
			if (parasiteNode.isLeaf())
				originalParasiteLeaves.add((InputParasiteNode) parasiteNode);
		}

		// create trees
		if (isBetaSplitModel) {
			if (tree == ViewerConstants.TREE_SELECTION_BOTH) {
				createRandomTreeByBetaSplitting(cophylogeny,
						originalHostLeaves.size(),
						ViewerConstants.TREE_SELECTION_HOST);
				createRandomTreeByBetaSplitting(cophylogeny,
						originalParasiteLeaves.size(),
						ViewerConstants.TREE_SELECTION_PARASITE);
			} else if (tree == ViewerConstants.TREE_SELECTION_HOST) {
				createRandomTreeByBetaSplitting(cophylogeny,
						originalHostLeaves.size(),
						ViewerConstants.TREE_SELECTION_HOST);
				recursiveCreateTreeCopy(cophylogeny.getParasiteRoot(),
						originalCophylogeny.getParasiteRoot(),
						ViewerConstants.TREE_SELECTION_PARASITE);
			} else if (tree == ViewerConstants.TREE_SELECTION_PARASITE) {
				recursiveCreateTreeCopy(cophylogeny.getHostRoot(),
						originalCophylogeny.getHostRoot(),
						ViewerConstants.TREE_SELECTION_HOST);
				createRandomTreeByBetaSplitting(cophylogeny,
						originalParasiteLeaves.size(),
						ViewerConstants.TREE_SELECTION_PARASITE);
			} else if (tree == ViewerConstants.TREE_SELECTION_NONE) {
				recursiveCreateTreeCopy(cophylogeny.getHostRoot(),
						originalCophylogeny.getHostRoot(),
						ViewerConstants.TREE_SELECTION_HOST);
				recursiveCreateTreeCopy(cophylogeny.getParasiteRoot(),
						originalCophylogeny.getParasiteRoot(),
						ViewerConstants.TREE_SELECTION_PARASITE);
			}
		} else if (isAgeModel) {
			if (tree == ViewerConstants.TREE_SELECTION_BOTH) {
				createRandomTreeByAge(cophylogeny, originalHostLeaves.size(),
						ViewerConstants.TREE_SELECTION_HOST);
				createRandomTreeByAge(cophylogeny,
						originalParasiteLeaves.size(),
						ViewerConstants.TREE_SELECTION_PARASITE);
			} else if (tree == ViewerConstants.TREE_SELECTION_HOST) {
				createRandomTreeByAge(cophylogeny, originalHostLeaves.size(),
						ViewerConstants.TREE_SELECTION_HOST);
				recursiveCreateTreeCopy(cophylogeny.getParasiteRoot(),
						originalCophylogeny.getParasiteRoot(),
						ViewerConstants.TREE_SELECTION_PARASITE);
			} else if (tree == ViewerConstants.TREE_SELECTION_PARASITE) {
				recursiveCreateTreeCopy(cophylogeny.getHostRoot(),
						originalCophylogeny.getHostRoot(),
						ViewerConstants.TREE_SELECTION_HOST);
				createRandomTreeByAge(cophylogeny,
						originalParasiteLeaves.size(),
						ViewerConstants.TREE_SELECTION_PARASITE);
			} else if (tree == ViewerConstants.TREE_SELECTION_NONE) {
				recursiveCreateTreeCopy(cophylogeny.getHostRoot(),
						originalCophylogeny.getHostRoot(),
						ViewerConstants.TREE_SELECTION_HOST);
				recursiveCreateTreeCopy(cophylogeny.getParasiteRoot(),
						originalCophylogeny.getParasiteRoot(),
						ViewerConstants.TREE_SELECTION_PARASITE);
			}
		}

		cophylogeny.getHostRoot().renumberNodes(0);
		if (tree == ViewerConstants.TREE_SELECTION_BOTH
				|| tree == ViewerConstants.TREE_SELECTION_HOST)
			cophylogeny.getHostRoot().renameNodes();
		cophylogeny.getParasiteRoot().renumberNodes(0);
		if (tree == ViewerConstants.TREE_SELECTION_BOTH
				|| tree == ViewerConstants.TREE_SELECTION_PARASITE)
			cophylogeny.getParasiteRoot().renameNodes();

		// create associations
		if (tree == ViewerConstants.TREE_SELECTION_BOTH
				|| tree == ViewerConstants.TREE_SELECTION_HOST
				|| tree == ViewerConstants.TREE_SELECTION_PARASITE)
			AssociationGenerator.createOriginalAssociations(originalHostLeaves,
					originalParasiteLeaves, hostLeaves, parasiteLeaves);
		else if (tree == ViewerConstants.TREE_SELECTION_NONE)
			AssociationGenerator
					.createRandomAssociationsWithOriginalCharacteristics(
							originalHostLeaves, originalParasiteLeaves,
							hostLeaves, parasiteLeaves, this.random);

		// create ranks
		createRandomRanks(cophylogeny, tree);

		return cophylogeny;
	}

	private void createRandomRanks(InputCophylogeny cophylogeny, int tree) {
		if (tree == ViewerConstants.TREE_SELECTION_HOST) {
			createRandomHostRanks(cophylogeny);
		} else if (tree == ViewerConstants.TREE_SELECTION_PARASITE) {
			List<IPositionNode> hostNodes = cophylogeny.getNodeList(cophylogeny
					.getHostRoot());
			int maxRank = 0;
			for (IPositionNode hostNode : hostNodes) {
				maxRank = Math.max(maxRank, ((IHostNode) hostNode).getRank());
			}
			createRandomParasiteRanks(cophylogeny, maxRank);
		} else if (tree == ViewerConstants.TREE_SELECTION_BOTH) {
			int maxRank = createRandomHostRanks(cophylogeny);
			createRandomParasiteRanks(cophylogeny, maxRank);
		}
	}

	private int createRandomHostRanks(InputCophylogeny cophylogeny) {
		int rank = 0;
		ArrayList<Tupel<BigDecimal, InputHostNode>> hostList = new ArrayList<Tupel<BigDecimal, InputHostNode>>();
		hostList.add(new Tupel<BigDecimal, InputHostNode>(BigDecimal.ONE,
				cophylogeny.getHostRoot()));
		while (!hostList.isEmpty()) {
			double rand = random.nextDouble();
			Tupel<BigDecimal, InputHostNode> hostTupel = null;
			for (Tupel<BigDecimal, InputHostNode> tupel : hostList) {
				rand -= tupel.getValue1().doubleValue();
				if (rand <= 0d) {
					hostTupel = tupel;
					break;
				}
			}
			if (hostTupel == null) {
				hostTupel = hostList.get(hostList.size() - 1);
			}
			InputHostNode node = hostTupel.getValue2();
			node.setRank(rank);
			rank++;
			hostList.remove(hostTupel);
			if (node.isLeaf()) {
				for (Tupel<BigDecimal, InputHostNode> tupel : hostList) {
					BigDecimal probabilityFraction = tupel
							.getValue1()

							.multiply(hostTupel.getValue1())
							.divide((BigDecimal.ONE.add(hostTupel.getValue1()
									.negate())),
									BaseSolverConstants.BIGDECIMAL_SCALE,
									RoundingMode.CEILING);
					tupel.setValue1(tupel.getValue1().add(probabilityFraction));
				}
			} else {

				BigDecimal fraction = hostTupel.getValue1().divide(
						BigDecimal.valueOf(node.getDirectChildrenCount()),
						BaseSolverConstants.BIGDECIMAL_SCALE,
						RoundingMode.CEILING);
				for (InputHostNode child : node.getChildren()) {
					hostList.add(new Tupel<BigDecimal, InputHostNode>(fraction,
							child));
				}
			}
		}
		return rank - 1;
	}

	private void createRandomParasiteRanks(InputCophylogeny cophylogeny,
			int maxRank) {
		List<IPositionNode> parasiteNodes = cophylogeny.getNodeList(cophylogeny
				.getParasiteRoot());
		for (int i = parasiteNodes.size() - 1; i >= 0; i--) {
			IParasiteNode parasiteNode = (IParasiteNode) parasiteNodes.get(i);
			ParasiteRank rank = null;
			if (!parasiteNode.getAssociations().isEmpty()) {
				for (IHostNode hostNode : parasiteNode.getAssociations()) {
					if (rank == null) {
						rank = new ParasiteRank(
								hostNode.getParent() == null ? 0 : hostNode
										.getParent().getRank(),
								hostNode.getRank());
					} else {
						rank.setRankFrom(Math.min(rank.getRankFrom(), hostNode
								.getParent() == null ? 0 : hostNode.getParent()
								.getRank()));
						rank.setRankTo(Math.max(rank.getRankTo(),
								hostNode.getRank()));
					}
				}
			}
			if (rank == null) {
				rank = new ParasiteRank(0, maxRank);
			}
			for (IParasiteNode child : parasiteNode.getChildren()) {
				rank.setRankTo(Math.min(rank.getRankTo(), child.getRank()
						.getRankTo()));
				rank.setRankFrom(Math.min(rank.getRankFrom(), child.getRank()
						.getRankFrom()));
			}
			parasiteNode.setRank(rank);
		}
	}

	private void recursiveCreateTreeCopy(InputNode node,
			InputNode originalNode, int tree) {
		node.setLabel(originalNode.getLabel());
		if (tree == ViewerConstants.TREE_SELECTION_HOST) {
			((InputHostNode) node).setRank(((InputHostNode) originalNode)
					.getRank());
		} else if (tree == ViewerConstants.TREE_SELECTION_PARASITE) {
			((InputParasiteNode) node).setRank(new ParasiteRank(
					((InputParasiteNode) originalNode).getRank()));
		}
		if (originalNode.getDirectChildrenCount() == 0) {
			if (tree == ViewerConstants.TREE_SELECTION_HOST) {
				this.hostLeaves.add((InputHostNode) node);
				// ((InputHostNode) node).setRank(((InputHostNode) originalNode)
				// .getRank());
			} else if (tree == ViewerConstants.TREE_SELECTION_PARASITE) {
				this.parasiteLeaves.add((InputParasiteNode) node);
				// ((InputParasiteNode) node).setRank(new ParasiteRank(
				// ((InputParasiteNode) originalNode).getRank()));
			}
			return;
		}

		for (InputNode originalChildNode : originalNode.getChildren()) {
			InputNode childNode = null;
			if (tree == ViewerConstants.TREE_SELECTION_HOST) {
				childNode = new InputHostNode();
			} else if (tree == ViewerConstants.TREE_SELECTION_PARASITE) {
				childNode = new InputParasiteNode();
			}
			node.addChild(childNode);
			recursiveCreateTreeCopy(childNode, originalChildNode, tree);
		}

	}

	private void createRandomTreeByBetaSplitting(InputCophylogeny cophylogeny,
			int leaves, int tree) {
		if (leaves < 2) {
			if (tree == ViewerConstants.TREE_SELECTION_HOST)
				this.hostLeaves.add(cophylogeny.getHostRoot());
			else if (tree == ViewerConstants.TREE_SELECTION_PARASITE)
				this.parasiteLeaves.add(cophylogeny.getParasiteRoot());
			return;
		}

		// create an ArrayList which contains the probability distribution q_n
		List<Double> q_n = getProbabilityDistributionForBetaSplitting(leaves);

		// get the splitting point by sum up until a probability
		int i = getRandomSplittingPointFromProbabilityDistribution(q_n, leaves);

		InputNode node = null;

		if (tree == ViewerConstants.TREE_SELECTION_HOST)
			node = cophylogeny.getHostRoot();
		else if (tree == ViewerConstants.TREE_SELECTION_PARASITE)
			node = cophylogeny.getParasiteRoot();

		recurseCreateByBetaSplitting(node, i, leaves - i, tree);
	}

	private void recurseCreateByBetaSplitting(InputNode node, int leftLeaves,
			int rightLeaves, int tree) {
		if (leftLeaves == 0 || rightLeaves == 0) {
			if (tree == ViewerConstants.TREE_SELECTION_HOST)
				this.hostLeaves.add((InputHostNode) node);
			else if (tree == ViewerConstants.TREE_SELECTION_PARASITE)
				this.parasiteLeaves.add((InputParasiteNode) node);
			return;
		}

		int i;
		InputNode lNode = null;
		InputNode rNode = null;
		if (tree == ViewerConstants.TREE_SELECTION_HOST) {
			lNode = new InputHostNode();
			rNode = new InputHostNode();
		} else if (tree == ViewerConstants.TREE_SELECTION_PARASITE) {
			lNode = new InputParasiteNode();
			rNode = new InputParasiteNode();
		}
		node.addChild(lNode);
		lNode.setParent(node);
		node.addChild(rNode);
		rNode.setParent(node);
		if (leftLeaves == 1)
			i = 1;
		else {
			// create a List which contains the probability distribution q_n
			List<Double> q_n = getProbabilityDistributionForBetaSplitting(leftLeaves - 1);
			// get the splitting point by sum up until a probability
			i = getRandomSplittingPointFromProbabilityDistribution(q_n,
					leftLeaves - 1);
		}
		recurseCreateByBetaSplitting(lNode, i, leftLeaves - i, tree);

		if (rightLeaves == 1)
			i = 1;
		else {
			// create a List which contains the probability distribution q_n
			List<Double> q_n = getProbabilityDistributionForBetaSplitting(rightLeaves - 1);
			// get the splitting point by sum up until a probability
			i = getRandomSplittingPointFromProbabilityDistribution(q_n,
					rightLeaves - 1);
		}
		recurseCreateByBetaSplitting(rNode, i, rightLeaves - i, tree);
	}

	private int getRandomSplittingPointFromProbabilityDistribution(
			List<Double> q_n, int n) {
		double x = this.random.nextDouble();
		int i;
		double prefixsum = 0.0;
		// prefixsum until x
		for (i = 1; i <= n - 1; i++) {
			double q_n_i;
			if (i <= (int) n / 2)
				q_n_i = ((Double) q_n.get(i - 1)).doubleValue();
			else
				q_n_i = ((Double) q_n.get(n - i - 1)).doubleValue();
			prefixsum = prefixsum + q_n_i;
			if (prefixsum > x)
				break;
		}
		return i;
	}

	private List<Double> getProbabilityDistributionForBetaSplitting(int n) {
		if (n < 1)
			return null;
		// check whether probability distribution is calculated already
		if (this.probabilityDistributions.containsKey(new Integer(n)))
			return this.probabilityDistributions.get(new Integer(n));

		List<Double> q_n = new ArrayList<Double>(n - 1);
		for (int i = 1; i <= n / 2; i++) {
			if (this.model == StatisticConstants.MODEL_MARKOV) {
				q_n.add(Double.valueOf(1d / ((double) (n - 1))));
			} else if (this.model == StatisticConstants.MODEL_ALDOUS) {
				q_n.add(Double.valueOf((1d / (2d * h(n - 1)))
						* ((double) n / ((double) (i * (n - i))))));
			} else if (this.model == StatisticConstants.MODEL_PDA) {
				q_n.add(Double.valueOf(0.5d * n_over_k(n, i)
						* c_component(n, i)));
			}

		}
		this.probabilityDistributions.put(new Integer(n), q_n);
		return q_n;
	}

	private double h(int n) {
		BigDecimal h = BigDecimal.ZERO;
		for (int i = 1; i <= n; i++) {

			h = h.add(BigDecimal.ONE.divide(BigDecimal.valueOf(i),
					BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING));

		}
		return h.doubleValue();
	}

	private double n_over_k(int n, int k) {
		BigDecimal result = BigDecimal.ONE;
		int dividend = n;
		int larger_divisor = Math.max(k, n - k);
		int smaler_divisor = Math.min(k, n - k);
		while (dividend > larger_divisor) {
			result = result
					.multiply(BigDecimal.valueOf(dividend).divide(
							BigDecimal.valueOf(smaler_divisor),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			dividend--;
			smaler_divisor--;
		}
		return result.doubleValue();
	}

	private double c_component(int n, int i) {
		BigDecimal result = BigDecimal.ONE;
		int divisor = c_n(n);
		int larger_dividend = Math.max(c_n(i), c_n(n - i));
		int smaler_dividend = Math.min(c_n(i), c_n(n - i));
		while (divisor > larger_dividend) {
			result = result.multiply(BigDecimal.valueOf(smaler_dividend)
					.divide(BigDecimal.valueOf(divisor),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			divisor -= 2;
			if (smaler_dividend > 1)
				smaler_dividend -= 2;
		}
		return result.doubleValue();
	}

	private int c_n(int n) {
		return Math.max(2 * n - 3, 1);
	}

	// --------------------------------------------------------------------------------
	// --------------------------------------------------------------------------------

	/**
	 * author: Stephanie Keller-Schmidt
	 * 
	 * @param cophylogeny
	 * @param leaves
	 */
	private void createRandomTreeByAge(InputCophylogeny cophylogeny,
			int leaves, int tree) {

		int time = 1;
		if (tree == ViewerConstants.TREE_SELECTION_HOST) {
			cophylogeny.getHostRoot().setRank(time);
			this.hostLeaves.add(cophylogeny.getHostRoot());
			recursiveCreateAgeModelTree(cophylogeny.getHostRoot(), tree, time,
					leaves, 1, cophylogeny);
		}

		else if (tree == ViewerConstants.TREE_SELECTION_PARASITE) {
			cophylogeny.getParasiteRoot().setRank(new ParasiteRank(time, time));
			this.parasiteLeaves.add(cophylogeny.getParasiteRoot());
			recursiveCreateAgeModelTree(cophylogeny.getParasiteRoot(), tree,
					time, leaves, 1, cophylogeny);
		} else {
			createRandomTreeByAge(cophylogeny, leaves,
					ViewerConstants.TREE_SELECTION_HOST);
			createRandomTreeByAge(cophylogeny, leaves,
					ViewerConstants.TREE_SELECTION_PARASITE);
		}

	}// end createRandomTreeByBetaSplitting

	/**
	 * author: Stephanie Keller-Schmidt
	 * 
	 * @param node
	 *            current node
	 * @param leavelist
	 *            list of leaves from current tree
	 * @param tree
	 *            defines if host or parasitenode is choosen
	 * @param maxLeafNumber
	 * @param currHostLeafNumber
	 * @param cophylogeny
	 */
	private void recursiveCreateAgeModelTree(InputNode node, int tree,
			int time, int leaves, int currentLeafNumber,
			InputCophylogeny cophylogeny) {

		InputNode lNode = null;
		InputNode rNode = null;

		// do HOSTSPECIATION
		if (tree == ViewerConstants.TREE_SELECTION_HOST) {
			lNode = new InputHostNode();
			rNode = new InputHostNode();

			((InputHostNode) node).setRank(time);
			((InputHostNode) lNode).setRank(time);
			((InputHostNode) rNode).setRank(time);

			node.addChild(lNode);
			lNode.setParent(node);
			node.addChild(rNode);
			rNode.setParent(node);

			int index = this.hostLeaves.indexOf(node);
			this.hostLeaves.add(index, (InputHostNode) rNode);
			this.hostLeaves.add(index, (InputHostNode) lNode);
			this.hostLeaves.remove(node);

			currentLeafNumber++;

			// preprocessing for next speciation step
			if (currentLeafNumber < leaves) {
				time++;
				// choose next speciation node
				double q_n[] = getProbabilityDistributionForAgeDependentSplitting(
						this.hostLeaves, time);
				InputNode node4split = (InputNode) getRandomSplittingPointFromProbabilityDistribution(
						q_n, this.hostLeaves);

				recursiveCreateAgeModelTree(node4split, tree, time, leaves,
						currentLeafNumber, cophylogeny);
			} else {
				for (InputHostNode hostLeaf : this.hostLeaves)
					hostLeaf.setRank(time);
			}

		}// end hostspeciation

		// do PARASITESPECIATION
		else if (tree == ViewerConstants.TREE_SELECTION_PARASITE) {
			lNode = new InputParasiteNode();
			rNode = new InputParasiteNode();

			((InputParasiteNode) node).setRank(new ParasiteRank(time, time));
			((InputParasiteNode) lNode).setRank(new ParasiteRank(time, time));
			((InputParasiteNode) rNode).setRank(new ParasiteRank(time, time));

			node.addChild(lNode);
			lNode.setParent(node);
			node.addChild(rNode);
			rNode.setParent(node);

			int index = this.parasiteLeaves.indexOf(node);
			this.parasiteLeaves.add(index, (InputParasiteNode) rNode);
			this.parasiteLeaves.add(index, (InputParasiteNode) lNode);
			this.parasiteLeaves.remove(node);

			currentLeafNumber++;

			// preprocessing for next speciation step
			if (currentLeafNumber < leaves) {
				time++;
				// choose next speciation node
				double q_n[] = getProbabilityDistributionForAgeDependentSplitting(
						this.parasiteLeaves, time);
				InputNode node4split = (InputNode) getRandomSplittingPointFromProbabilityDistribution(
						q_n, this.parasiteLeaves);

				recursiveCreateAgeModelTree(node4split, tree, time, leaves,
						currentLeafNumber, cophylogeny);
			} else {
				for (InputParasiteNode parasiteLeaf : this.parasiteLeaves)
					parasiteLeaf.setRank(new ParasiteRank(time, time));
			}
		}// end parasitespeciation

	}// end recursiveCreateAgeModelTree

	/**
	 * author: Stephanie Keller-Schmidt
	 * 
	 * getProbabilityDistributionForBetaSplitting(int n) including age
	 * 
	 * @param leavelist
	 * @param time
	 * @return
	 */
	private double[] getProbabilityDistributionForAgeDependentSplitting(
			List<? extends InputNode> leavelist, int time) {

		double q_n[] = new double[leavelist.size()];

		double sum = 0d;
		Iterator<? extends InputNode> itIn = leavelist.iterator();

		while (itIn.hasNext()) {

			InputNode node = itIn.next();
			// which kind of node??? -> typeOf
			if (node instanceof InputHostNode) {
				sum += ((1d / (double) (time - ((InputHostNode) node).getRank())));
			} else {
				sum += (1d / (double) (time - ((InputParasiteNode) node)
						.getRank().getRankFrom()));
			}
		}

		int i = 0;
		Iterator<? extends InputNode> it = leavelist.iterator();

		InputNode n = null;
		while (it.hasNext()) {
			n = it.next();

			if (n instanceof InputHostNode) {
				if (i == 0) {
					q_n[i] = getAgeDependentProbHost(time, (InputHostNode) n,
							sum);
				} else {
					q_n[i] = (q_n[i - 1] + getAgeDependentProbHost(time,
							(InputHostNode) n, sum));
				}
			} else {
				if (i == 0) {
					q_n[i] = getAgeDependentProbParasite(time,
							(InputParasiteNode) n, sum);
				} else {
					q_n[i] = (q_n[i - 1] + getAgeDependentProbParasite(time,
							(InputParasiteNode) n, sum));
				}
			}
			i++;

		}

		return q_n;
	}// end getProbabilityDistributionForAgeDependentSplitting

	/**
	 * author: Stephanie Keller-Schmidt
	 * 
	 * getProbabilityDistributionForBetaSplitting(int n) including age
	 * 
	 * @param leavelist
	 * @return
	 */
	@SuppressWarnings("unused")
	private double[] getEqualProbabilityDistribution(
			List<? extends InputNode> leavelist) {

		double q_n[] = new double[leavelist.size()];

		for (int i = 0; i < leavelist.size(); i++) {
			if (i == 0) {
				q_n[i] = (1d / (double) leavelist.size());
			} else {
				q_n[i] = q_n[i - 1] + (1d / (double) leavelist.size());
			}
			i++;
		}

		return q_n;
	}// end getEqualProbabilityDistribution

	/**
	 * author: Stephanie Keller-Schmidt
	 * 
	 * @param q_n
	 * @param leavelist
	 * @return
	 */
	private InputNode getRandomSplittingPointFromProbabilityDistribution(
			double[] q_n, List<? extends InputNode> leavelist) {

		double x = this.random.nextDouble();
		int i = 0;

		while (i < q_n.length - 1 && q_n[i] < x) {
			i++;
		}
		return leavelist.get(i);
	}// end getRandomSplittingPointFromProbabilityDistribution

	/**
	 * 
	 * @param time
	 * @param n
	 * @param norm
	 */
	private double getAgeDependentProbHost(int time, InputHostNode n,
			double norm) {
		return ((1d / (double) (time - (n.getRank()))) / norm);
	}// end getAgeDependentProbHost

	/**
	 * 
	 * @param time
	 * @param n
	 * @param norm
	 */
	private double getAgeDependentProbParasite(int time, InputParasiteNode n,
			double norm) {
		return (((1d / (double) (time - n.getRank().getRankFrom()))) / norm);
	}// end getAgeDependentProbParasite

}
